﻿using System.Collections;
using System.Collections.Generic;
using System.IO;
using UnityEngine;

public class ForestViewManager : MonoBehaviour
{
    [SerializeField]
    private Tree treePrefab;

    [SerializeField]
    private float interval;

    [SerializeField]
    private Transform anchor;

    [SerializeField]
    private float range;

    [SerializeField]
    private float unitTreeSize;

    [SerializeField]
    private Camera cameraTarget;

    [SerializeField]
    private float showDuration;

    [SerializeField]
    private Color tintColor;

    [SerializeField]
    private AnimationCurve curve;

    [SerializeField]
    private TreeDataSet treeDataSet;

    [Space(5)]
    [SerializeField]
    private ForestViewUI ui;
    
    private List<Score> scoreList = new List<Score>();

    private object lockObj = new object();
    private bool canSpawn = false;

    private float timer;

    public Score focusScore;

    private float showTimer;
    private Vector3 startPoint;
    private Vector3 targetPoint;
    private float targetOrthoSize;

    void Start()
    {
        if (ui != null)
        {
            ui.gameObject.SetActive(false);
        }
        
        ScoreManager.onScoreUpdated += OnScoreUpdated;
        ScoreManager.instance.LoadScores();

        var json = PlayerPrefs.GetString("score", null);
        Debug.Log(json);

        if (!string.IsNullOrEmpty(json))
        {
            focusScore = JsonUtility.FromJson<Score>(json);
        }
        else
        {
            focusScore = new Score();
            focusScore.name = "Test";
            focusScore.score = 500;

        }

        // Debug.Log("Score: " + focusScore.score);

        startPoint = cameraTarget.transform.position;
        targetOrthoSize = Mathf.Clamp(10f * focusScore.score / unitTreeSize / 3, 10, 30);
        var targetY = targetOrthoSize * 0.33f;
        targetPoint = new Vector3(0, targetY, startPoint.z);

        StartCoroutine(ShowCoroutine());
    }

    void OnDestroy()
    {
        ScoreManager.onScoreUpdated -= OnScoreUpdated;
    }

    private IEnumerator ShowCoroutine()
    {
        showTimer = 0;
        while (showTimer < showDuration)
        {
            cameraTarget.transform.position = Vector3.Lerp(startPoint, targetPoint, curve.Evaluate(showTimer / showDuration));
            cameraTarget.orthographicSize = Mathf.Lerp(10, targetOrthoSize, curve.Evaluate(showTimer / showDuration));
            showTimer += Time.deltaTime;

            yield return new WaitForEndOfFrame();
        }

        SpawnTrees();

        var pos = anchor.position + new Vector3(0, treeDataSet.data[focusScore.treeId % treeDataSet.data.Length].spawnOffset);
        var tree = Instantiate(treePrefab, pos, Quaternion.identity);
        tree.SetHeight(Mathf.Max(0.5f, focusScore.score / unitTreeSize));

        tree.LoadData(treeDataSet.data[focusScore.treeId % treeDataSet.data.Length]);

        if (ui != null)
        {
            ui.gameObject.SetActive(true);
        }
    }

    private void OnScoreUpdated(Score score)
    {
        lock (lockObj)
        {
            var index = scoreList.BinarySearch(score);
            if (index < 0)
            {
                scoreList.Insert(~index, score);
            }
        }
    }

    public void SpawnTrees()
    {
        timer = interval;
        canSpawn = true;
    }

    void Update()
    {
        if (Input.GetKeyDown(KeyCode.P)) SpawnTrees();

        if (!canSpawn || !ScoreManager.instance.scoreInitialized) return;

        timer -= Time.deltaTime;
        if (timer <= 0)
        {
            SpawnTree();
            timer = interval;
        }
    }

    private void SpawnTree()
    {
        lock (lockObj)
        {
            if (scoreList.Count <= 0) return;

            var pos = anchor.position + new Vector3((float)scoreList[0].pos, treeDataSet.data[scoreList[0].treeId % treeDataSet.data.Length].spawnOffset);
            var tree = Instantiate(treePrefab, pos, Quaternion.identity);
            tree.SetHeight(Mathf.Max(0.5f, scoreList[0].score / unitTreeSize));
            var index = Random.Range(-32, -52);
            tree.SetColor(Color.Lerp(Color.white, tintColor, Mathf.InverseLerp(-2, -52, index)));
            tree.SetSortingIndex(index);
            // tree.LoadData(treeDataSet.data[Random.Range(0, treeDataSet.data.Length)]);
            tree.LoadData(treeDataSet.data[scoreList[0].treeId % treeDataSet.data.Length]);

            scoreList.RemoveAt(0);
        }
    }
}
