﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class PlayerMovement : MonoBehaviour
{
    [SerializeField]
    private Transform spriteTransform;

    [SerializeField]
    private float jitter;
    [SerializeField]
    private float jitterLerp;
    [SerializeField]
    private float dieDecayDuration;

    [SerializeField]
    private GameEvent hazardEnteredEvent;
    [SerializeField]
    private GameEvent hazardExitedEvent;

    [SerializeField]
    private PlayerInputSystem inputSource;

    public float vSpeed;
    public float hSpeed;

    private Vector3 currentPosition;
    private Vector3 currentJitterOffset;
    private Quaternion[] normals;
    private Vector3[] jitterEndpoints;

    public float speedFactor = 1.0f;
    
    private Player parentPlayer = null;

    public bool isDying { get; private set; }

    private float dieDecayTimer;

    public bool hasStopped { get { return isDying && dieDecayTimer <= 0;} }

    void Start()
    {
        dieDecayTimer = dieDecayDuration;
        normals = new Quaternion[] { Quaternion.Euler(0, 0, 90f), Quaternion.Euler(0, 0, -90f) };
    }

    public void Init(Player player)
    {
        parentPlayer = player;
    }

    // Update is called once per frame
    void Update()
    {
        if (parentPlayer != null && !parentPlayer.IsAlive)
        {
            return;
        }

        if (isDying)
        {
            dieDecayTimer = Mathf.Max(dieDecayTimer - Time.deltaTime, 0);
        }
        
        var totalSpeedFactor = speedFactor * dieDecayTimer / dieDecayDuration;
        
        var h = inputSource.GetInput();

        currentPosition.y = -vSpeed * totalSpeedFactor;
        currentPosition.x = h * hSpeed * totalSpeedFactor;

        currentPosition = Vector3.ClampMagnitude(currentPosition, vSpeed * totalSpeedFactor);

        jitterEndpoints = new Vector3[] { normals[0] * currentPosition.normalized * jitter, normals[1] * currentPosition.normalized * jitter };
        currentJitterOffset = Vector3.Lerp(jitterEndpoints[0], jitterEndpoints[1], Random.value) * totalSpeedFactor;
        spriteTransform.localPosition = Vector2.Lerp(spriteTransform.localPosition, currentJitterOffset, jitterLerp);

        transform.position += currentPosition * Time.deltaTime;

    }

    void OnTriggerEnter2D(Collider2D collider)
    {
        Hazard hazard = collider.GetComponent<Hazard>();
        if(hazard)
        {
            speedFactor = hazard.Slowdown() * dieDecayTimer / dieDecayDuration;
            hazardEnteredEvent.Raise();
            GameManager.instance.Player.AddDecayDelta(hazard.Decay());
        }
    }
    
    void OnTriggerExit2D(Collider2D collider)
    {
        Hazard hazard = collider.GetComponent<Hazard>();
        if(hazard)
        {
            speedFactor = hazard.Revert() * dieDecayTimer / dieDecayDuration;
            hazardExitedEvent.Raise();
            GameManager.instance.Player.AddDecayDelta(-hazard.Decay());
        }
    }

    public void StartDecay()
    {
        dieDecayTimer = dieDecayDuration;
        isDying = true;
    }
    
    public void StopDecay()
    {
        isDying = false;
        dieDecayTimer = dieDecayDuration;
    }
}
